==> hypercrud.server-0.2.0-20170821.200246-4.jar: hypercrud/server/datomic/root_init.clj <==
(ns hypercrud.server.datomic.root-init
  (:require [clojure.java.io :as io]
            [datomic.api :as d]
            [hypercrud.util.core :as util]))


(defn generate-db-record [db-name]
  {:db/id (d/tempid :db.part/user)
   ; todo https://tools.ietf.org/html/rfc3986#section-2
   :database/ident db-name})

(defn reflect-schema [conn]
  (let [$ (d/db conn)]
    (->> (d/q '[:find [?a ...] :in $ :where [?a :db/ident]] $)
         (mapv #(d/touch (d/entity $ %)))
         ;filter out datomic attributes, todo this is a huge hack
         (filter #(> (:db/id %) 62)))))

(defn attr->hc-attr [attr]
  (->> attr
       (into {})
       (util/map-keys (fn [ident]
                        (condp = ident
                          :db/id ident
                          (keyword "attribute" (name ident)))))))

; root isn't parameterized that well
; the current hf-src will create a database for root AND hyperfiddle AND hyperfiddle-users
; this means 3 databases on top of existing infra :(
(defn init
  ([transactor-uri hf-schema]
   (let [hf-data
         ; the only reason this root db-record exists in this loader is to satisfy init-datomic
         ; it can be deleted when hypercrud.server.db-root/root-id is
         [(generate-db-record "root")]
         ]
     (init transactor-uri hf-schema hf-data)))
  ([transactor-uri hf-schema hf-data]
   (let [root-transactor-uri (str transactor-uri "root")]
     (if-not (d/create-database root-transactor-uri)
       (throw (new Error "Cannot initialize from existing database"))

       (let [root-conn (d/connect root-transactor-uri)]
         (println "Installing root schema")
         @(d/transact root-conn hf-schema)

         (when hf-data
           (println "Installing hyperfiddle data")
           @(d/transact root-conn hf-data))

         ; all of the following reflection shouldn't be necessary after removing the root
         ; user fiddles should execute it on demand
         (println "Reflecting existing user databases")
         (let [db-names (->> (d/get-database-names (str transactor-uri "*"))
                             (remove #(= root-transactor-uri (str transactor-uri %))))]

           (println "Registering user databases")
           @(d/transact root-conn (mapv generate-db-record db-names))

           (println "Reflecting user schemas")
           ; this will fail if attributes already exist
           @(d/transact root-conn (->> db-names
                                       (mapcat #(reflect-schema (d/connect (str transactor-uri %))))
                                       (mapv attr->hc-attr)
                                       (into #{})
                                       (into []))))

         (println "Finished initializing"))))))

(defn init-from-file
  ([transactor-uri hf-schema-src]
   (assert (.exists (io/file hf-schema-src)))
   (init-from-file transactor-uri hf-schema-src nil))
  ([transactor-uri hf-schema-src hf-data-src]
   (assert (.exists (io/file hf-schema-src)))
   (init transactor-uri
         (-> hf-schema-src slurp read-string)
         (some-> hf-data-src slurp read-string))))

==> hypercrud.server-0.2.0-20170821.234817-5.jar: hypercrud/server/datomic/root_init.clj <==
(ns hypercrud.server.datomic.root-init
  (:require [clojure.java.io :as io]
            [datomic.api :as d]
            [hypercrud.util.core :as util]))


(defn generate-db-record [db-name]
  {:db/id (d/tempid :db.part/user)
   ; todo https://tools.ietf.org/html/rfc3986#section-2
   :database/ident db-name})

(defn reflect-schema [conn]
  (let [$ (d/db conn)]
    (->> (d/q '[:find [?a ...] :in $ :where [?a :db/ident]] $)
         (mapv #(d/touch (d/entity $ %)))
         ;filter out datomic attributes, todo this is a huge hack
         (filter #(> (:db/id %) 62)))))

(defn attr->hc-attr [attr]
  (->> attr
       (into {})
       (util/map-keys (fn [ident]
                        (condp = ident
                          :db/id ident
                          (keyword "attribute" (name ident)))))))

(defn reflect-hc-attributes [transactor-uri db-name]
  (->> (reflect-schema (d/connect (str transactor-uri db-name)))
       (mapv attr->hc-attr)
       (into #{})
       (into [])))

(defn initialize-user-schemas [root-conn transactor-uri db-names]
  ; this will fail if attributes already exist
  (doseq [db-name db-names]
    (println (str "Reflecting schema: " db-name))
    @(d/transact root-conn (reflect-hc-attributes transactor-uri db-name))))

; root isn't parameterized that well
; the current hf-src will create a database for root AND hyperfiddle AND hyperfiddle-users
; this means 3 databases on top of existing infra :(
(defn init
  ([transactor-uri hf-schema]
   (let [hf-data
         ; the only reason this root db-record exists in this loader is to satisfy init-datomic
         ; it can be deleted when hypercrud.server.db-root/root-id is
         [(generate-db-record "root")]
         ]
     (init transactor-uri hf-schema hf-data)))
  ([transactor-uri hf-schema hf-data]
   (let [root-transactor-uri (str transactor-uri "root")]
     (if-not (d/create-database root-transactor-uri)
       (throw (new Error "Cannot initialize from existing database"))

       (let [root-conn (d/connect root-transactor-uri)]
         (println "Installing root schema")
         @(d/transact root-conn hf-schema)

         (when hf-data
           (println "Installing hyperfiddle data")
           @(d/transact root-conn hf-data))

         (println "Reflecting existing user databases")
         (doseq [db-name (->> (d/get-database-names (str transactor-uri "*"))
                              (remove #(= root-transactor-uri (str transactor-uri %))))]
           (println (str "Registering db: " db-name))
           @(d/transact root-conn [(generate-db-record db-name)]))

         (println "Finished initializing"))))))

(defn init-from-file
  ([transactor-uri hf-schema-src]
   (assert (.exists (io/file hf-schema-src)))
   (init-from-file transactor-uri hf-schema-src nil))
  ([transactor-uri hf-schema-src hf-data-src]
   (assert (.exists (io/file hf-schema-src)))
   (init transactor-uri
         (-> hf-schema-src slurp read-string)
         (some-> hf-data-src slurp read-string))))
